"""Implements the enStratus Cloud API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class Subscription(Resource):
    PATH = 'geography/Subscription'
    COLLECTION_NAME = 'subscriptions'
    PRIMARY_KEY = 'region_id'

    def __init__(self, region_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__region_id = region_id

    @property
    def region_id(self):
        return self.__region_id

    @classmethod
    def region(cls, region_id, **kwargs):
        """Returns subscription for given Region"""

        r = Resource(cls.PATH+"/"+str(region_id))
        r.request_details = 'basic'
        s = r.get()
        return s

    @classmethod
    def all(cls, **kwargs):
        """Returns subscriptions for all regions"""

        r = Resource(cls.PATH)
        r.request_details = 'basic'
        s = r.get()
        return s


"""Implements the enStratus Cloud API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class Cloud(Resource):
    PATH = 'geography/Cloud'
    COLLECTION_NAME = 'clouds'
    PRIMARY_KEY = 'cloud_id'

    def __init__(self, cloud_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__cloud_id = cloud_id

    @property
    def cloud_id(self):
        """`int` - The unique enStratus id for this cloud"""
        return self.__cloud_id

    @lazy_property
    def cloud_provider_console_url(self):
        """`str` - URL of the cloud provider's own console"""
        return self.__cloud_provider_console_url

    @lazy_property
    def cloud_provider_logo_url(self):
        """`str` - enStratus installation URL to this cloud's logo"""
        return self.__cloud_provider_logo_url

    @lazy_property
    def cloud_provider_name(self):
        """`str` - The name of the vendor providing the cloud"""
        return self.__cloud_provider_name

    @lazy_property
    def compute_account_number_label(self):
        """`str` - enStratus localized label key for the cloud's user account number"""
        return self.__compute_account_number_label

    @lazy_property
    def compute_access_key_label(self):
        """`str` - enStratus localized label key for the cloud's API key"""
        return self.__compute_access_key_label

    @lazy_property
    def compute_x509_cert_label(self):
        """`str` - enStratus localized label key for the cloud's X509 certificate"""
        return self.__compute_x509_cert_label

    @lazy_property
    def compute_x509_key_label(self):
        """`str` - enStratus localized label key for the cloud's X509 key"""
        return self.__compute_x509_key_label

    @lazy_property
    def compute_delegate(self):
        """`str` - Dasein Cloud API Java class for interacting with this cloud"""
        return self.__compute_delegate

    @lazy_property
    def compute_endpoint(self):
        """`str` - Comma-separated list of cloud API endpoints"""
        return self.__compute_endpoint

    @lazy_property
    def compute_secret_key_label(self):
        """`str` - enStratus localized label key for the cloud's API secret key"""
        return self.__compute_secret_key_label

    @lazy_property
    def documentation_label(self):
        """`str` - enStratus localized label key"""
        return self.__documentation_label

    @lazy_property
    def name(self):
        """`str` - The enStratus name of the cloud"""
        return self.__name

    @lazy_property
    def private_cloud(self):
        """`bool` - If the cloud is a private cloud or not"""
        return self.__private_cloud

    @lazy_property
    def status(self):
        """`str` - Status of the cloud in enStratus"""
        return self.__status

    @classmethod
    def all(cls, keys_only=False, **kwargs):
        """Return all clouds

        :param keys_only: Return :attr:`cloud_id` instead of :class:`Cloud`
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :class:`Cloud` or :attr:`cloud_id`
        """
        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {}

        if 'public_only' in kwargs:
            params['publicOnly'] = kwargs['public_only']
        if 'status' in kwargs:
            params['status'] = kwargs['status']

        c = r.get(params=params)

        if r.last_error is None:
            if keys_only is True:
                return [i['cloudId'] for i in c[cls.COLLECTION_NAME]]
            else:
                clouds = []
                for i in c[cls.COLLECTION_NAME]:
                    cloud = cls(i['cloudId'])
                    if 'detail' in kwargs:
                        cloud.request_details = kwargs['detail']
                    cloud.params = params
                    cloud.load()
                    clouds.append(cloud)
                return clouds
        else:
            return r.last_error



from mixcoatl.resource import Resource
from mixcoatl.admin.job import Job
from mixcoatl.utils import camel_keys
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.decorators.lazy import lazy_property

import json, sys, time

class RelationalDatabase(Resource):
    """A relational database is a database as a service offering supporting a relational model."""

    PATH = 'platform/RelationalDatabase'
    COLLECTION_NAME = 'relational_databases'
    PRIMARY_KEY = 'relational_database_id'

    def __init__(self, relational_database_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__relational_database_id = relational_database_id

    @property
    def relational_database_id(self):
        """`int` - The DCM ID of this relational_database"""
        return self.__relational_database_id

    @lazy_property
    def label(self):
        """`str` - The label assigned to the relational_database"""
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def account(self):
        """`int` - The DCM account under which the relational database is registered."""
        return self.__account

    @account.setter
    def account(self, a):
        self.__account = a

    @lazy_property
    def admin_password(self):
        """`str` - The password of an admin user of the database."""
        return self.__admin_password

    @admin_password.setter
    def admin_password(self, p):
        self.__admin_password = p

    @lazy_property
    def admin_user(self):
        """`str` - The username of an admin user of the database."""
        return self.__admin_user

    @admin_user.setter
    def admin_user(self, u):
        self.__admin_user = u

    @lazy_property
    def allocated_storage_in_gb(self):
        """`int` - The amount of storage allocated to the relational database in GB."""
        return self.__allocated_storage_in_gb

    @allocated_storage_in_gb.setter
    def allocated_storage_in_gb(self, g):
        self.__allocated_storage_in_gb = g

    @lazy_property
    def backup_window(self):
        """`str` - The time windows when automated snapshot backups happen."""
        return self.__backup_window

    @lazy_property
    def cloud(self):
        """`dict` - The cloud provided where the instance is located."""
        return self.__cloud

    @lazy_property
    def created_timestamp(self):
        """`str` - Creation timestamp of the database."""
        return self.__created_timestamp

    @lazy_property
    def current_recovery_point(self):
        """`str` - Specifies the latest time to which the database can be restored with point-in-time restore."""
        return self.__current_recovery_point

    @lazy_property
    def customer(self):
        """`dict` - The customer account for the relational_database."""
        return self.__customer

    @lazy_property
    def data_center(self):
        """`dict` - The specific datacenter where the instance is located."""
        return self.__data_center

    @data_center.setter
    def data_center(self, d):
        self.__data_center = {u'data_center_id': d}

    @lazy_property
    def description(self):
        """`str` - The description of the relational_database"""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def dns_name(self):
        """`str` - A DNS name associated with the relational database server."""
        return self.__dns_name

    @description.setter
    def dns_name(self, d):
        self.__dns_name = d

    @lazy_property
    def engine(self):
        """ `enum` - The database engine powering this relational database. examples: MYSQL, ORACLE11G"""
        return self.__engine

    @engine.setter
    def engine(self, e):
        self.__engine = e

    @lazy_property
    def ha(self):
        """`bool` - Indicates if high availability is enabled."""
        return self.__ha

    @ha.setter
    def ha(self, h):
        self.__ha = h

    @lazy_property
    def maintenance_window(self):
        """`str` - The time window when maintenance is allowed to happen."""
        return self.__maintenance_window

    @lazy_property
    def name(self):
        """`str` - A user-friendly name to describe this relational database."""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def owning_group(self):
        """`str` - The group who has ownership over this relational database."""
        return self.__owning_group

    @lazy_property
    def owning_user(self):
        """`str` - The user who is the owner of this relational database."""
        return self.__owning_user

    @lazy_property
    def port(self):
        """`int` - The port number the database is running on."""
        return self.__port

    @port.setter
    def port(self, n):
        self.__port = n

    @lazy_property
    def provider_id(self):
        """`str` - How this relational database is identified by cloud provider."""
        return self.__provider_id

    @lazy_property
    def rdbms_product(self):
        """`int` - The relational database product associated with this database."""
        return self.__rdbms_product

    @rdbms_product.setter
    def rdbms_product(self, r):
        self.__rdbms_product = r

    @lazy_property
    def region(self):
        """`dict` - The region this database is running in."""
        return self.__region

    @region.setter
    def region(self, r):
        self.__region = r

    @lazy_property
    def status(self):
        """`str` - The current status of the database *(i.e. `RUNNING` or `PAUSED`)*."""
        return self.__status

    def reload(self):
        """Reload resource data from API calls"""
        if self.relational_database_id is not None:
            self.load()
        elif self.current_job is None:
            self.load()
        else:
            if Job.wait_for(self.current_job):
                job = Job(self.current_job)
                self.__relational_database_id = job.message
                self.load()
            else:
                return self.last_error

    @required_attrs(['relational_database_id'])
    def destroy(self, reason='no reason provided'):
        """Terminate relational_database instance with reason :attr:`reason`

        :param reason: The reason for terminating the relational_database
        :type reason: str.
        :returns: bool -- Result of API call
        """
        p = self.PATH+"/"+str(self.relational_database_id)
        qopts = {'reason':reason}
        return self.delete(p, params=qopts)

    @required_attrs(['name', 'description', 'port', 'engine', 'rdbms_product', 'data_center',
                     'allocated_storage_in_gb', 'admin_user', 'admin_password'])
    def launch(self, callback=None):
        """Launches a relational database with the configured parameters

        >>> def cb(j): print(j)
        >>> r = RelationalDatabase()
        >>> r.name = 'relational_database-1-test'
        >>> r.description = 'my first database launch'
        >>> r.port = 3306
        >>> r.engine = 'MYSQL'
        >>> r.rdbms_product = ''
        >>> r.data_center = 3451
        >>> r.allocated_storage_in_gb = 5
        >>> r.admin_user = 'root'
        >>> r.admin_password = 'adminpassword'
        >>> r.launch(callback=cb)

        :param callback: Optional callback to send the results of the API call
        :type callback: func.
        :returns: int -- The job id of the launch request
        :raises: :class:`RelationalDatabaseLaunchException`, :class:`mixcoatl.decorators.validations.ValidationException`
        """
#        optional_attrs = ['vlan', 'firewalls', 'keypair', 'label', 'cmAccount', 'environment', 'cm_scripts', 'p_scripts', 'volumeConfiguration']
#        if self.relational_database_id is not None:
#            raise RelationalDatabaseLaunchException('Cannot launch an already running relational_database: %s' % self.relational_database_id)
#
#        payload = {'launch':
#                    [{
#                        'productId': self.provider_product_id,
#                        'budget': self.budget,
#                        'machineImage': camel_keys(self.machine_image),
#                        'description': self.description,
#                        'name': self.name,
#                        'dataCenter': camel_keys(self.data_center)
#                    }]}
#
#        for oa in optional_attrs:
#            try:
#                if getattr(self, oa) is not None:
#                    if oa == 'cm_scripts':
#                        payload['launch'][0].update({'scripts':getattr(self, oa)})
#                    elif oa == 'p_scripts':
#                        payload['launch'][0].update({'personalities':getattr(self, oa)})
#                    elif oa == 'volumeConfiguration':
#                        payload['launch'][0].update({'volumeConfiguration':{u'raidlevel':'RAID0', u'volumeCount':1, u'volumeSize':2, u'fileSystem':'ext3', u'mountPoint':'/mnt/data'}})
#                    elif oa == 'vlan':
#                        payload['launch'][0].update({'vlan':camel_keys(getattr(self, oa))})
#                    else:
#                        payload['launch'][0].update({oa:getattr(self, oa)})
#            except AttributeError:
#                pass

        self.post(data=json.dumps(payload))
        if self.last_error is None:
            if callback is not None:
                callback(self.current_job)
            else:
                return self.current_job
        else:
            raise RelationalDatabaseLaunchException(self.last_error)

    def wait_for(self, status='RUNNING', callback = None):
        """Blocks execution until the current relational_database has status of :attr:`status`

        :param status: The status to expect before continuing *(i.e. `RUNNING` or `PAUSED`)*
        :type status: str.
        :param callback: Optional callback to be called with the final :class:`RelationalDatabase`` when ``status`` is reached
        :type callback: func.
        :raises: `RelationalDatabaseException`
        """
        if self.relational_database_id is None:
            raise RelationalDatabaseException('Must be called with an existing relational_database.')
        initial_status = self.status
        if self.last_error is None:
            if initial_status == status:
                return self
            while self.status != status:
                time.sleep(5)
                self.load()
                if self.last_error is not None:
                    raise RelationalDatabaseException(self.last_error)
                else:
                    continue
            if callback is not None:
                callback(self)
            else:
                return self

    @classmethod
    def all(cls, **kwargs):
        """Get a list of all known relational_databases

        >>> RelationalDatabase.all()
        [{'relational_database_id':1,...},{'relational_database_id':2,...}]

        :returns: list -- a list of :class:`RelationalDatabase`
        :raises: RelationalDatabaseException
        """
        r = Resource(cls.PATH)
        r.request_details = 'basic'

        if 'params' in kwargs:
          params = kwargs['params']
        else:
          params = []
        s = r.get(params=params)
        if r.last_error is None:
            relational_databases = [cls(relational_database['relationalDatabaseId']) for relational_database in s[cls.COLLECTION_NAME]]
            return relational_databases
        else:
            raise RelationalDatabaseException(r.last_error)

class RelationalDatabaseException(BaseException): pass
class RelationalDatabaseLaunchException(RelationalDatabaseException): pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
from mixcoatl.admin.job import Job

import json
import time

class LaunchConfiguration(Resource):
    PATH = 'automation/LaunchConfiguration'
    COLLECTION_NAME = 'launchConfigurations'
    PRIMARY_KEY = 'launch_configuration_id'

    def __init__(self, launch_configuration_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__launch_configuration_id = launch_configuration_id

    @property
    def launch_configuration_id(self):
        return self.__launch_configuration_id

    @property
    def deployment(self):
        return self.__deployment

    @property
    def array_volume_capacity(self):
        return self.__array_volume_capacity

    @property
    def array_volume_count(self):
        return self.__array_volume_count

    @lazy_property
    def label(self):
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def customer(self):
        return self.__customer

    @lazy_property
    def created_timestamp(self):
        return self.__created_timestamp

    @lazy_property
    def created(self):
        return self.__created

    @lazy_property
    def cm_account(self):
        return self.__cm_account

    @lazy_property
    def firewalls(self):
        return self.__firewalls

    @firewalls.setter
    def firewalls(self, f):
        self.__firewalls = f

    @lazy_property
    def network(self):
        return self.__network

    @network.setter
    def network(self, n):
        self.__network = n

    @lazy_property
    def personalities(self):
        return self.__personalities

    @personalities.setter
    def personalities(self, p):
        self.__personalities = p

    @lazy_property
    def primary_machine_image(self):
        return self.__primary_machine_image

    @primary_machine_image.setter
    def primary_machine_image(self, p):
        self.__primary_machine_image = p

    @lazy_property
    def primary_product(self):
        return self.__primary_product

    @primary_product.setter
    def primary_product(self, p):
        self.__primary_product = p

    @lazy_property
    def raid_level(self):
        return self.__raid_level

    @raid_level.setter
    def raid_level(self, r):
        self.__raid_level = r

    @lazy_property
    def recovery_delay_in_minutes(self):
        return self.__recovery_delay_in_minutes

    @recovery_delay_in_minutes.setter
    def recovery_delay_in_minutes(self, d):
        self.__recovery_delay_in_minutes = d

    @lazy_property
    def region(self):
        return self.__region

    @region.setter
    def region(self, r):
        self.__region = r

    @lazy_property
    def scripts(self):
        return self.__scripts

    @scripts.setter
    def scripts(self, s):
        self.__scripts = s

    @lazy_property
    def secondary_machine_image(self):
        return self.__secondary_machine_image

    @secondary_machine_image.setter
    def secondary_machine_image(self, s):
        self.__secondary_machine_image = s

    @lazy_property
    def secondary_product(self):
        return self.__secondary_product

    @secondary_product.setter
    def secondary_product(self, s):
        self.__secondary_product = s

    @lazy_property
    def server_name_template(self):
        return self.__server_name_template

    @server_name_template.setter
    def server_name_template(self, l):
        self.__server_name_template = l

    @lazy_property
    def server_type(self):
        return self.__server_type

    @server_type.setter
    def server_type(self, s):
        self.__server_type = s

    @lazy_property
    def snapshot_interval_in_minutes(self):
        return self.__snapshot_interval_in_minutes

    @snapshot_interval_in_minutes.setter
    def snapshot_interval_in_minutes(self, l):
        self.__snapshot_interval_in_minutes = l

    @lazy_property
    def subnet(self):
        return self.__subnet

    @subnet.setter
    def subnet(self, s):
        self.__subnet = s

    @property
    def tier(self):
        return self.__tier

    @tier.setter
    def tier(self, t):
        self.__tier = t

    @lazy_property
    def use_encrypted_volumes(self):
        return self.__use_encrypted_volumes

    @use_encrypted_volumes.setter
    def use_encrypted_volumes(self, u):
        self.__use_encrypted_volumes = u

    @lazy_property
    def use_hypervisor_stats(self):
        return self.__use_hypervisor_stats

    @use_hypervisor_stats.setter
    def use_hypervisor_stats(self, l):
        self.__use_hypervisor_stats = l

    @lazy_property
    def removable(self):
        return self.__removable

    @lazy_property
    def status(self):
        return self.__status

    @classmethod
    def all(cls, **kwargs):
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return r.last_error

    @required_attrs(['tier',
										 'region',
										 'primary_machine_image',
										 'primary_product_id'
										 ])

    def create(self, callback=None):
        """Creates a new launch configuration

        :param callback: Optional callback to send the resulting :class:`Job`
        :raises: :class:`TierCreationException`
        """
        #lc.name = name
        #lc.region = region
        #lc.primary_product = primary_product_id
        #lc.secondary_product = secondary_product_id
        #lc.primary_machine_image = primary_machine_image
        #lc.secondary_machine_image = secondary_machine_image
        #lc.server_name_template = server_name_template
        #lc.tier = tier
        #lc.firewall = firewall
        #lc.region = region

        parms = [{'tier':{'tierId':self.tier},
                 'primaryMachineImage':{'machineImageId':self.primary_machine_image},
                 'primaryProduct':{'productId':self.primary_product_id},
                 'firewalls':[{'firewallId':self.firewalls}],
                 'region':{'regionId':self.region}}]

        payload = {'addLaunchConfiguration':camel_keys(parms)}

        response=self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            print response
            return response
        else:
            raise LaunchConfigurationCreationException(self.last_error)

class LaunchConfigurationException(BaseException): pass

class LaunchConfigurationCreationException(LaunchConfigurationException):
    """Launch Configuration Creation Exception"""
    pass



"""Implements the enStratus ServerProduct API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property

class ServerProduct(Resource):
    PATH = 'infrastructure/ServerProduct'
    COLLECTION_NAME = 'serverProducts'
    PRIMARY_KEY = 'product_id'

    def __init__(self, product_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__product_id = product_id

    @property
    def product_id(self):
        """`int` - The unique enStratus id for this product"""
        return self.__product_id

    @lazy_property
    def cloud(self):
        """`dict` - The cloud in which this server product is valid"""
        return self.__cloud

    @lazy_property
    def architecture(self):
        """`str` - The CPU architecture of the underlying vCPUs"""
        return self.__architecture

    @lazy_property
    def cpu_count(self):
        """`int` - The number of vCPUs for this product"""
        return self.__cpu_count

    @lazy_property
    def cpu_speed_in_mhz(self):
        """`int` - The speed of the vCPUs for this product"""
        return self.__cpu_speed_in_mhz

    @lazy_property
    def currency(self):
        """`str` - The currenct in which the associated costs are charged"""
        return self.__currency

    @lazy_property
    def description(self):
        """`str` - A detailed description of what this product represents"""
        return self.__description

    @lazy_property
    def disk_size_in_gb(self):
        """`int` - The size of the root volume in GB"""
        return self.__disk_size_in_gb

    @lazy_property
    def hourly_rate(self):
        """`int` - The hourly rate for usage of this product"""
        return self.__hourly_rate

    @lazy_property
    def name(self):
        """`str` - The user-friendly name of this product"""
        return self.__name

    @lazy_property
    def platform(self):
        """`str` - The operating system for this product"""
        return self.__platform

    @lazy_property
    def provider_product_id(self):
        """`str` - The cloud provider's unique id for this product"""
        return self.__provider_product_id

    @lazy_property
    def provider_region_id(self):
        """`str` - The cloud provider region id where this product is valid"""
        return self.__provider_region_id

    @lazy_property
    def ram_in_mb(self):
        """`int` - The memory size for this product in MB"""
        return self.__ram_in_mb

    @lazy_property
    def software(self):
        """`str` - Any additional software installed on this product"""
        return self.__software

    @classmethod
    def all(cls, region_id, **kwargs):
        """Return all server products

        :param region_id: The region id to search in
        :type region_id: int.
        :param keys_only: Return :attr:`product_id` or :class:`ServerProduct`
        :type keys_only: bool.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :attr:`product_id` or :class:`ServerProduct`
        :raises: :class:`ServerProductException`
        """
        from mixcoatl.utils import uncamel_keys
        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {'regionId':region_id}
        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        c = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                products = [item['productId'] for item in c[cls.COLLECTION_NAME]]
            else:
                products = []
                for i in c[cls.COLLECTION_NAME]:
                    product = cls(i['productId'])
                    if 'detail' in kwargs:
                        product.request_details = kwargs['detail']
                    product.load()
                    products.append(product)
            return products
        else:
            raise ServerProductException(r.last_error)

class ServerProductException(BaseException): pass



"""
mixcoatl.admin.user
-------------------

Implements access to the enStratus User API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
import json
import time

class User(Resource):
    """A user within the enStratus environment"""

    PATH = 'admin/User'
    COLLECTION_NAME = 'users'
    PRIMARY_KEY = 'user_id'

    def __init__(self, user_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__user_id = user_id

    @property
    def user_id(self):
        """`int` The enStratus unique id of this user across all customer accounts"""
        return self.__user_id

    @lazy_property
    def account(self):
        """`dict` The enStratus account"""
        return self.__account

    @account.setter
    def account(self, a):
        # pylint: disable-msg=C0111,W0201
        self.__account = a

    @lazy_property
    def account_user_id(self):
        """`str` or `None` A unique identifier to reference this user's access
            to a specific account
        """
        return self.__account_user_id

    @lazy_property
    def alpha_name(self):
        """`str` The user's full name in the form of `Last name, First name`"""
        return self.__alpha_name

    @lazy_property
    def billing_codes(self):
        """`list` The billing codes against which this user has provisioning rights

        .. note::

            Only present when this user is queried in the context of an `account_id`

        """
        return self.__billing_codes

    @billing_codes.setter
    def billing_codes(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__billing_codes = b

    @lazy_property
    def cloud_console_password(self):
        """`str` The encrypted password that the user can log into the underlying cloud with"""
        return self.__cloud_console_password

    @lazy_property
    def cloud_api_public_key(self):
        """`str` The encrypted public key the user can make API calls to the underlying cloud with"""
        return self.__cloud_api_public_key

    @lazy_property
    def cloud_api_secret_key(self):
        """`str` The encrypted secret key the user can make API calls to the underlying cloud with"""
        return self.__cloud_api_secret_key

    @lazy_property
    def customer(self):
        """`dict` The customer record to which this user belongs"""
        return self.__customer

    @lazy_property
    def editable(self):
        """`bool` Indicates if the core values of this user may be changed"""
        return self.__editable

    @lazy_property
    def email(self):
        """`str` Email is a unique identifier that enables a given user to
            identify themselves to enStratus
        """
        return self.__email

    @email.setter
    def email(self, e):
        # pylint: disable-msg=C0111,W0201
        self.__email = e

    @lazy_property
    def family_name(self):
        """`str` The family name of the user"""
        return self.__family_name

    @family_name.setter
    def family_name(self, f):
        # pylint: disable-msg=C0111,W0201
        self.__family_name = f

    @lazy_property
    def given_name(self):
        """`str` The given name of the user"""
        return self.__given_name

    @given_name.setter
    def given_name(self, g):
        # pylint: disable-msg=C0111,W0201
        self.__given_name = g

    @lazy_property
    def groups(self):
        """`list` The group membership of this user idependent of any individual accounts"""
        return self.__groups

    @groups.setter
    def groups(self, g):
        # pylint: disable-msg=C0111,W0201
        self.__groups = g

    @lazy_property
    def has_cloud_api_access(self):
        """`bool` Indicates that the user has access to the underlying cloud API (i.e. AWS IAM)"""
        return self.__has_cloud_api_access

    @lazy_property
    def has_cloud_console_access(self):
        """`bool` Indicates that the user has access to the underlying cloud console (i.e. AWS IAM)"""
        return self.__has_cloud_console_access

    @lazy_property
    def legacy_user_id(self):
        """`int` The legacy user id"""
        return self.__legacy_user_id

    @lazy_property
    def notifications_targets(self):
        """`dict` The various targets configured for delivery of notifications"""
        return self.__notification_targets

    @lazy_property
    def notifications_settings(self):
        """`dict` Notification settings configured for this user"""
        return self.__notification_settings

    @lazy_property
    def status(self):
        """`str` The current status of this user in enStratus"""
        return self.__status

    @lazy_property
    def time_zone(self):
        """`str` The timezone id for the user's prefered time zone"""
        return self.__time_zone

    @lazy_property
    def vm_login_id(self):
        """`str` The username the user will use to login to cloud instances
            for shell or remote desktop access
        """
        return self.__vm_login_id

    @lazy_property
    def ssh_public_key(self):
        """`str` The public key to grant the user access to Unix instances"""
        return self.__ssh_public_key

    @lazy_property
    def password(self):
        """`str` DCM login password"""
        return self.__password

    @password.setter
    def password(self, p):
        self.__password = p

    @required_attrs(['user_id'])
    def grant(self, account_id, groups, billing_codes):
        """Grants the user access to the specified account. :attr:`reason`

        :param account_id: Account ID of the account to grant access.
        :type account_id: int.
        :param groups: List of group ID the user will belong to.
        :type groups: list.
        :param billing_codes: List of billing code the user will use.
        :type billing_codes: list.
        :returns: bool -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.user_id))

        group_list = []
        billing_code_list = []
        for group in groups:
            group_list.append({"groupId": group})
        for billing_code in billing_codes:
            billing_code_list.append({"billingCodeId": billing_code})

        payload = {"grant":[{"account": {"accountId": account_id},
                             "groups": group_list,
                             "billingCodes": billing_code_list}]}

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['account', 'given_name', 'family_name', 'email', 'groups', 'billing_codes'])
    def create(self):
        """Creates a new user."""

        billing_code_list = []
        for billing_code in self.billing_codes:
            billing_code_list.append({"billingCodeId": billing_code})

        parms = [{'givenName':self.given_name,
                  'familyName': self.family_name,
                  'email': self.email,
                  'groups': [{'groupId':self.groups}],
                  'account': {'accountId':self.account},
                  'billingCodes': billing_code_list}]

        if self.password is not None:
            parms[0].update({'password': self.password})

        payload = {'addUser':camel_keys(parms)}

        response=self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise UserCreationException(self.last_error)

    @classmethod
    def all(cls, keys_only=False, **kwargs):
        """Return all users

        .. note::

            The keys used to make the request determine results visibility

        :param keys_only: Return :attr:`user_id` instead of :class:`User`
        :type keys_only: bool.
        :param detail: str. The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :class:`User` or :attr:`user_id`
        :raises: :class:`UserException`
        """
        r = Resource(cls.PATH)
        if 'detail' in kwargs:
            r.request_details = kwargs['detail']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i['userId'] for i in x[cls.COLLECTION_NAME]]
            else:
                return [cls(i['userId']) for i in x[cls.COLLECTION_NAME]]
        else:
            raise UserException(r.last_error)

class UserException(BaseException): pass

class UserCreationException(UserException):
    """User Creation Exception"""
    pass



"""
mixcoatl.analytics.server_analytics
-----------------------------------

Implements access to the enStratus ServerAnalytics API

.. note::

    Unlike other objects, ServerAnalytics only applies to a single server.
    There is no concept of 'all analytics across all servers'

"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class ServerAnalytics(Resource):
    """Server analytics represent the performance of an individual server over
        a specified period of time

    :param server_id: The server represented in the analytics
    :type server_id: int.
    :param interval: The interval in minutes for the requested data points
    :type interval: int.
    :param period_start: The start time in UNIX milliseconds for the first datapoint
    :param period_start: int.
    :param period_end: The end time in UNIX milliseconds for the last datapoint
    :param period_end: int.
    """

    PATH = 'analytics/ServerAnalytics'
    COLLECTION_NAME = 'analytics'
    PRIMARY_KEY = 'server_id'

    def __init__(self, server_id=None, **kwargs):
        Resource.__init__(self, kwargs)
        self.__server_id = server_id

    @property
    def server_id(self):
        """`int` The unique enStratus id for the server in the request"""
        return self.__server_id

    @lazy_property
    def data_points(self):
        """`list` Data points representing a snapshot of the server's state
            at a given interval point
        """
        return self.__data_points

    @lazy_property
    def period_start(self):
        """`str` The actual date and time when the data set begins"""
        return self.__period_start

    @lazy_property
    def period_end(self):
        """`str` The date and time for the last data point in the response"""
        return self.__period_end

    @lazy_property
    def interval_in_minutes(self):
        """`int` The interval between data points in minutes delivered in this response"""
        return self.__interval_in_minutes

    @classmethod
    def all(cls, server_id, data_only=False, **kwargs):
        """Get all analytics for `server_id`

        :param server_id: The server represented in the analytics
        :type server_id: int.
        :param data_only: Return only the datapoints
        :type data_only: bool.
        :param interval: The interval in minutes for the requested data points
        :type interval: int.
        :param period_start: The start time in UNIX milliseconds for the first datapoint
        :param period_start: int.
        :param period_end: The end time in UNIX milliseconds for the last datapoint
        :param period_end: int.
        :returns: :class:`ServerAnalytics` or `list` of :attr:`data_points`
        """

        params = {}
        if 'interval' in kwargs:
            params['interval'] = kwargs['interval']
        if 'period_end' in kwargs:
            params['periodEnd'] = kwargs['period_end']
        if 'period_start' in kwargs:
            params['periodStart'] = kwargs['period_start']
        s = cls(server_id)
        if 'details' in kwargs:
            s.request_details = kwargs['details']
        else:
            s.request_details = 'basic'
        s.params = params
        s.load()
        if s.last_error is None:
            if data_only is True:
                return s.data_points
            else:
                return s
        else:
            raise ServerAnalyticsException(s.last_error)

class ServerAnalyticsException(BaseException): pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
import json

class ConfigurationManagementService(Resource):
    PATH = 'automation/ConfigurationManagementService'
    COLLECTION_NAME = 'cmServices'
    PRIMARY_KEY = 'cm_system_id'

    def __init__(self, cm_account_id=None, *args, **kwargs):
        Resource.__init__(self)

    @lazy_property
    def cm_system(self):
        return self.__cm_system

    @lazy_property
    def service_endpoint(self):
        return self.__service_endpoint

    @lazy_property
    def properties(self):
        return self.__properties
                
    @lazy_property
    def cm_service_id(self):
        return self.__cm_service_id

    @lazy_property
    def budget(self):
        return self.__budget

    @budget.setter
    def budget(self, b):
        self.__budget = b

    @lazy_property
    def removable(self):
        return self.__removable
                             
    @lazy_property
    def description(self):
        return self.__description

    @description.setter
    def description(self, b):
        self.__description = b

    @lazy_property
    def name(self):
        return self.__name

    @name.setter
    def name(self, b):
        self.__name = b

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def customer(self):
        return self.__customer

    @required_attrs(['budget', 'description', 'name', 'endpoint', 'cm_system_id'])
    def create(self):
        """Creates a new CM service."""

        parms = [{"budget": self.budget,
                  "serviceEndpoint": self.endpoint,
                  "description": self.description,
                  "name": self.name,
                  "label": "red",
                  "cmSystem": {"cmSystemID": self.cm_system_id}}]

        payload = {"addService": camel_keys(parms)}
        print json.dumps(payload)

	response = self.post(data=json.dumps(payload))
	if self.last_error is None:
		self.load()
		return response
	else:
		raise CMCreationException(self.last_error)
        
    @classmethod
    def all(cls, **kwargs):
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return x.last_error

class CMException(BaseException): pass
	
class CMCreationException(CMException):
    """CM Creation Exception"""
    pass


"""Common helper utilities for use with mixcoatl"""

def uncamel(val):
    """Return the snake case version of :attr:`str`

    >>> uncamel('deviceId')
    'device_id'
    >>> uncamel('dataCenterName')
    'data_center_name'
    """
    import re
    s = lambda val: re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', val).lower().strip('_')
    return s(val)

def uncamel_keys(d1):
    """Return :attr:`d1` with all keys converted to snake case

    >>> d = {'myThings':[{'thingId':1,'someThings':{'firstThing':'a_thing'}}]}
    >>> uncamel_keys(d)
    {'my_things': [{'thing_id': 1, 'some_things': {'first_thing': 'a_thing'}}]}
    """
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k, v in d1.iteritems():
        new_key = uncamel(k)
        if isinstance(v, dict):
            d2[new_key] = uncamel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [uncamel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2

def camelize(val):
    """Return the camel case version of a :attr:`str`

    >>> camelize('this_is_a_thing')
    'thisIsAThing'
    """
    s = ''.join([t.title() for t in val.split('_')])
    return s[0].lower()+s[1:]

def camel_keys(d1):
    """Return :attr:`d1` with all keys converted to camel case

    >>> b = {'my_things': [{'thing_id': 1, 'some_things': {'first_thing': 'a_thing'}}]}
    >>> camel_keys(b)
    {'myThings': [{'thingId': 1, 'someThings': {'firstThing': 'a_thing'}}]}
    """
    d2 = dict()
    if not isinstance(d1, dict):
        return d1
    for k, v in d1.iteritems():
        new_key = camelize(k)
        if isinstance(v, dict):
            d2[new_key] = camel_keys(v)
        elif isinstance(v, list):
            d2[new_key] = [camel_keys(item) for item in v]
        else:
            d2[new_key] = v
    return d2

def convert(val):
    """Return :attr:`input` converted from :class:`unicode` to :class:`str`

    >>> convert(u'bob')
    'bob'
    >>> convert([u'foo', u'bar'])
    ['foo', 'bar']
    >>> convert({u'foo':u'bar'})
    {'foo': 'bar'}
    """
    if isinstance(val, dict):
        return dict((convert(key), convert(value)) for key, value in val.iteritems())
    elif isinstance(val, list):
        return [convert(element) for element in val]
    elif isinstance(val, unicode):
        return val.encode('utf-8')
    else:
        return val



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.admin.job import Job
from mixcoatl.utils import camel_keys
from mixcoatl.utils import camelize

import json

class FirewallRule(Resource):
    PATH = 'network/FirewallRule'
    COLLECTION_NAME = 'rules'
    PRIMARY_KEY = 'firewall_rule_id'

    def __init__(self, firewall_rule_id = None, *args, **kwargs):
        if 'detail' in kwargs:
            self.request_details = kwargs['detail']

        Resource.__init__(self)
        self.__firewall_rule_id = firewall_rule_id

    @property
    def firewall_rule_id(self):
        """`int` - The unique enStratus id for this rule"""
        return self.__firewall_rule_id

    @lazy_property
    def direction(self):
        """`str` - The direction of the rule: `ingress` or `egress`"""
        return self.__direction

    @direction.setter
    def direction(self, d):
        self.__direction = d

    @lazy_property
    def firewall(self):
        """`dict` - The firewall to which this rule belongs"""
        return self.__firewall

    @firewall.setter
    def firewall(self, fwid):
        self.__firewall = {'firewall_id': fwid}

    @lazy_property
    def network_address(self):
        """`str` - The CIDR for the source or destination of traffic"""
        return self.__network_address

    @network_address.setter
    def network_address(self, addr):
        self.__network_address = addr

    @lazy_property
    def protocol(self):
        """`str` - The network protocol for this rule"""
        return self.__protocol

    @protocol.setter
    def protocol(self, p):
        self.__protocol = p

    @lazy_property
    def start_port(self):
        """`int` or `None` - The start port for this rule"""
        return self.__start_port

    @start_port.setter
    def start_port(self, p):
        self.__start_port = p

    @lazy_property
    def end_port(self):
        """`int` or `None` - The end port for this rule"""
        return self.__end_port

    @end_port.setter
    def end_port(self, p):
        self.__end_port = p

    @lazy_property
    def rule_provider_id(self):
        """`str` - Provider's firewall rule ID"""
        return self.__rule_provider_id

    @required_attrs(['firewall', 'network_address', 'protocol', 'direction', 'start_port', 'end_port'])
    def create(self, **kwargs):
        """Create a new firewall rule

            .. warning::

                Does not currently support adding ICMP rules

        :param reason: Reason for the new rule
        :type reason: str.
        :returns: `bool`
        :raises: :class:`FirewallRuleException`
        """

        if 'reason' not in kwargs:
            reason = 'Added by mixcoatl'
        else:
            reason = kwargs['reason']

        payload = {'add_rule':[{
            'firewall_id': self.firewall['firewall_id'],
            'direction': self.direction,
            'start_port': self.start_port,
            'end_port': self.end_port,
            'protocol': self.protocol,
            'reason': reason,
            'cidr' : self.network_address}]}

        if self.firewall_rule_id is not None:
            raise FirewallRuleException('Cannot modify existing firewall rule')

        self.post(self.PATH, data=json.dumps(camel_keys(payload)))

        if self.last_error is None:
            return True
        else:
            if self.last_request.status_code == 418:
                return True
            else:
                raise FirewallRuleException(self.last_error)

    @required_attrs(['firewall_rule_id'])
    def remove(self, reason):
        """Remove a firewall rule

        :param reason: Reason for removing the rule
        :type reason: str.
        :returns: `bool`
        :raises: :class:`FirewallRuleException`
        """

        params = {'reason': reason}
        self.delete(self.PATH+'/'+str(self.firewall_rule_id), params=params)

        if self.last_error is None:
            return True
        else:
            raise FirewallRuleException(self.last_error)

    @classmethod
    def all(cls, firewall_id, **kwargs):
        """List all rules for `firewall_id`

        :param firewall_id: The id of the firewall to list rules for
        :type firewall_id: int.
        :param detail: Level of detail to return - `basic` or `extended`
        :type detail: str.
        :param keys_only: Return only :attr:`firewall_rule_id` in results
        :type keys_only: bool.
        :returns: `list` of :attr:`firewall_rule_id` or :class:`FirewallRule`
        :raises: :class:`FirewallRuleException`
        """

        params = {}
        r = Resource(cls.PATH)
        r.request_details = 'none'
        if 'detail' in kwargs:
            request_details = kwargs['detail']
        else:
            request_details = 'extended'

        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        params['firewallId'] = firewall_id
        x = r.get(params=params)
        if r.last_error is None:
            keys = [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            if keys_only is True:
                rules = keys
            else:
                rules = []
                for i in x[cls.COLLECTION_NAME]:
                    key = i[camelize(cls.PRIMARY_KEY)]
                    rule = cls(key)
                    rule.request_details = request_details
                    rule.load()
                    rules.append(rule)
            return rules
        else:
            raise FirewallRuleException(r.last_error)

def add_rule(firewall_id, network, proto, direction, start, end, reason):
    """Add a firewall rule to a firewall

    >>> f = add_rule(136663, '10.1.1.1/32', 'TCP', 'INGRESS', 15000, 15000, 'inbound api')

    """
    f = FirewallRule()
    f.firewall = firewall_id
    f.network_address = network
    f.protocol = proto
    f.direction = direction
    f.start_port = start
    f.end_port = end
    return f.create(reason=reason)

def delete_rule(rule_id, reason='rule removed by mixcoatl'):
    """Remove a firewall rule

    :param rule_id: The id of the firewall rule to remove
    :type rule_id: int.
    :param reason: The reason for removing the rule
    :type reason: string
    :returns: `bool`
    :raises: :class:`FirewallRuleException`
    """
    f = FirewallRule(rule_id)
    return f.remove(reason)

class FirewallRuleException(BaseException):
    """Generic Exception for FirewallRules"""
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize

class Script(Resource):
    PATH = 'automation/Script'
    COLLECTION_NAME = 'scripts'
    PRIMARY_KEY = 'cmAccountId'

    def __init__(self, cmAccountId = None, *args, **kwargs):
        Resource.__init__(self)
        self.__cmAccountId = cmAccountId 

    @property
    def cmAccountId(self):
        return self.__cmAccountId

    @classmethod
    def all(cls, cmAccountId, **kwargs):
        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {'cmAccountId':cmAccountId}
        c = r.get(params=params)
        return c[cls.COLLECTION_NAME]

class ScriptException(BaseException): pass


import os, datetime

from mixcoatl.exceptions import ConfigException

def validate(date_text):
    try:
        datetime.datetime.strptime(date_text, '%Y-%m-%d')
        return date_text
    except ValueError:
        raise ValueError("Incorrect data format, should be YYYY-MM-DD")

class Config(object):
    # pylint: disable-msg=E0710
    def __init__(self):
        self.access_key = None
        self.secret_key = None
        self.endpoint = None
        self.user_agent = 'mixcoatl'
        self.api_version = None
        self.basepath = None
        self.default_api_version = '2012-06-15'
        self.ssl_verify = None

    def configure(self):
        if self.access_key is None:
            if 'ES_ACCESS_KEY' in os.environ:
                self.set_access_key(os.environ['ES_ACCESS_KEY'])
            else:
                raise ConfigException('missing ES_ACCESS_KEY')
        if self.secret_key is None:
            if 'ES_SECRET_KEY' in os.environ:
                self.set_secret_key(os.environ['ES_SECRET_KEY'])
            else:
                raise ConfigException('missing ES_SECRET_KEY')
        if self.api_version is None:
            if 'ES_API_VERSION' in os.environ:
                self.set_api_version(os.environ['ES_API_VERSION'])
            else:
                if 'ES_ENDPOINT' in os.environ:
                  str = os.environ['ES_ENDPOINT'].split('/')

                  if validate(str[-1]):
                    self.set_api_version(str[-1])
                  else:
                    self.set_api_version(self.default_api_version)
                else:
                  self.set_api_version(self.default_api_version)

            self.set_basepath('/api/enstratus/%s' % self.api_version)

        if self.endpoint is None:
            if 'ES_ENDPOINT' in os.environ:
                self.set_endpoint(os.environ['ES_ENDPOINT'])
            else:
                self.set_endpoint('https://api.enstratus.com'+self.basepath)

        if self.ssl_verify is None:
            if 'ES_SSL_VERIFY' in os.environ:
                self.set_ssl_verify(os.environ['ES_SSL_VERIFY'])
            else:
                self.set_ssl_verify('1')

    def set_access_key(self, key):
        self.access_key = key

    def set_secret_key(self, key):
        self.secret_key = key

    def set_endpoint(self, ep):
        self.endpoint = ep

    def set_api_version(self, version):
        self.api_version = version

    def set_basepath(self, basepath):
        self.basepath = basepath

    def set_ssl_verify(self, verify):
        if verify == '0':
            self.ssl_verify = False
        else:
            self.ssl_verify = True



"""Implements the enStratus Deployment API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
from mixcoatl.admin.job import Job

import json
import time

class Deployment(Resource):
    PATH = 'automation/Deployment'
    COLLECTION_NAME = 'deployments'
    PRIMARY_KEY = 'deployment_id'

    def __init__(self, deployment_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__deployment_id = deployment_id

    @property
    def deployment_id(self):
        return self.__deployment_id

    @lazy_property
    def backup_window(self):
        return self.__backup_window

    @lazy_property
    def budget(self):
        return self.__budget

    @budget.setter
    def budget(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__budget = b

    @lazy_property
    def label(self):
        return self.__label

    @lazy_property
    def creation_timestamp(self):
        return self.__creation_timestamp

    @lazy_property
    def customer(self):
        return self.__customer

    @lazy_property
    def description(self):
        return self.__description

    @description.setter
    def description(self, d):
        # pylint: disable-msg=C0111,W0201
        self.__description = d

    @lazy_property
    def for_service_catalog(self):
        return self.__for_service_catalog

    @lazy_property
    def launch_timestamp(self):
        return self.__launch_timestamp

    @lazy_property
    def maintenance_window(self):
        return self.__maintenance_window

    @lazy_property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        # pylint: disable-msg=C0111,W0201
        self.__name = n

    @lazy_property
    def owning_groups(self):
        return self.__owning_groups

    @lazy_property
    def owning_user(self):
        return self.__owning_user

    @lazy_property
    def regions(self):
        return self.__regions

    @regions.setter
    def regions(self, r):
        # pylint: disable-msg=C0111,W0201
        self.__regions = r

    @lazy_property
    def removable(self):
        return self.__removable

    @lazy_property
    def load_balancers(self):
        return self.__load_balancers

    @lazy_property
    def reason_not_removable(self):
        return self.__reason_not_removable

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def e_type(self):
        return self.__e_type

    @lazy_property
    def dr_storage(self):
        return self.__dr_storage

    @classmethod
    def all(cls, **kwargs):
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return r.last_error

    @required_attrs(['name', 'description','budget','region'])
    def create(self, callback=None):
        """Creates a new deployment

        :param callback: Optional callback to send the resulting :class:`Job`
        :raises: :class:`DeploymentCreationException`
        """

        parms = [{'budget': self.budget,
                    'regionId': self.region,
                    'description': self.description,
										'name': self.name}]


        payload = {'addDeployment':camel_keys(parms)}

        response=self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise DeploymentCreationException(self.last_error)

class DeploymentException(BaseException): pass

class DeploymentCreationException(DeploymentException):
    """Deployment Creation Exception"""
    pass



"""Implements the enStratus Region API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class Region(Resource):
    """A region is a logical sub-infrastructure within a cloud"""
    PATH = 'geography/Region'
    COLLECTION_NAME = 'regions'
    PRIMARY_KEY = 'region_id'

    def __init__(self, region_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__region_id = region_id

    @property
    def region_id(self):
        """`int` - The unique enStratus id for this region"""
        return self.__region_id

    @lazy_property
    def cloud(self):
        """`dict` - The cloud to which this region belongs"""
        return self.__cloud

    @lazy_property
    def customer(self):
        """`dict` - The customer to whom this region belongs"""
        return self.__customer

    @lazy_property
    def jurisdiction(self):
        """`str` - The legal context in which the region operates"""
        return self.__jurisdiction

    @lazy_property
    def name(self):
        """`str` - The user-friendly name for the region"""
        return self.__name

    @lazy_property
    def provider_id(self):
        """`str` - The cloud provider's unique id for the region"""
        return self.__provider_id

    @lazy_property
    def status(self):
        """`str` - The current status of the region"""
        return self.__status

    @lazy_property
    def description(self):
        """`str` - The description of the region"""
        return self.__description

    @classmethod
    def all(cls, **kwargs):
        """Return all regions

        :param account_id: Limit results to regions with the specified account
        :type account_id: int.
        :param jurisdiction: Limit results to the specified jurisdiction
        :type jurisdiction: str.
        :param scope: Limit results to `all` (Default - cross-cloud)
            or `account` (cloud-specific)
        :type scope: str.
        :param keys_only: Return :attr:`region_id` instead of :class:`Region`
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :class:`Region` or :attr:`region_id`
        :raises: :class:`RegionException`
        """
        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {}
        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False
        for x in ['account_id', 'jurisdiction', 'scope']:
            if x in kwargs:
                params[camelize(x)] = kwargs[x]
        c = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                regions = [item['regionId'] for item in c[cls.COLLECTION_NAME]]
            else:
                regions = []
                for i in c[cls.COLLECTION_NAME]:
                    region = cls(i['regionId'])
                    if 'detail' in kwargs:
                        region.request_details = kwargs['detail']
                    region.load()
                    regions.append(region)
            return regions
        else:
            raise RegionException(r.last_error)

class RegionException(BaseException): pass



"""Implements the enStratus DataCenter API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property

class DataCenter(Resource):
    """
    A data center is a part of a regional infrastructure that has some ability
        to share resources with other data centers in the region
    """
    PATH = 'geography/DataCenter'
    COLLECTION_NAME = 'dataCenters'
    PRIMARY_KEY = 'data_center_id'

    def __init__(self, data_center_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__data_center_id = data_center_id

    @property
    def data_center_id(self):
        """`int` - The unique enStratus id for this data center"""
        return self.__data_center_id

    @lazy_property
    def description(self):
        """`str` - A description of the data center"""
        return self.__description

    @lazy_property
    def name(self):
        """`str` - A user-friendly name for the data center"""
        return self.__name

    @lazy_property
    def provider_id(self):
        """`str` - The cloud provider's unique id for the data center"""
        return self.__provider_id

    @lazy_property
    def region(self):
        """`dict` = The region to which this data center belongs"""
        return self.__region

    @lazy_property
    def status(self):
        """`str` - The current status of the data center"""
        return self.__status

    @classmethod
    def all(cls, region_id, **kwargs):
        """Return all data centers

        :param region_id: Required. The region to query against
        :type region_id: int.
        :param keys_only: Return :attr:`data_center_id` instead of :class:`DataCenter`
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` of :class:`DataCenter` or :attr:`data_center_id`
        :raises: :class:`DataCenterException`
        """
        if 'keys_only' in kwargs:
            keys_only = kwargs['keys_only']
        else:
            keys_only = False

        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {'regionId':region_id}
        c = r.get(params=params)
        if r.last_error is None:
            if keys_only is True:
                dcs = [i['dataCenterId'] for i in c[cls.COLLECTION_NAME]]
            else:
                dcs = []
                for i in c[cls.COLLECTION_NAME]:
                    dc = cls(i['dataCenterId'])
                    if 'detail' in kwargs:
                        dc.request_details = kwargs['detail']
                    dc.load()
                    dcs.append(dc)
            return dcs
        else:
            raise DataCenterException(r.last_error)

class DataCenterException(BaseException): pass



import time
import base64
import hashlib
import hmac
from mixcoatl.settings.load_settings import settings

def get_sig(http_method, path):
    """Return a signature valid for use in making enStratus API calls

    :param http_method: The `http_method` used to make the API call
    :param path: The `path` used in the API call
    """
    timestamp = int(round(time.time() * 1000))
    signpath = settings.basepath+'/'+path

    parts = []
    parts.append(settings.access_key)
    parts.append(http_method)
    parts.append(signpath)
    parts.append(timestamp)
    parts.append(settings.user_agent)
    # pylint: disable-msg=E1101
    dm = hashlib.sha256
    to_sign = ':'.join([str(x) for x in parts])
    d = hmac.new(settings.secret_key, msg=to_sign, digestmod=dm).digest()
    b64auth = base64.b64encode(d).decode()
    return {
            'timestamp':timestamp,
            'signature':b64auth,
            'access_key':settings.access_key,
            'ua':settings.user_agent}



"""Implements the ability to list CM environments via the API"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
from mixcoatl.admin.job import Job

class Environment(Resource):
    PATH = 'automation/Environment'
    COLLECTION_NAME = 'environments'
    PRIMARY_KEY = 'environmentId'

    def __init__(self, environmentId=None, *args, **kwargs):
        Resource.__init__(self)
        self.__environmentId = environmentId

    @classmethod
    def all(cls, cmAccountId, **kwargs):
        r = Resource(cls.PATH)
        r.request_details = 'basic'
        params = {'cmAccountId':cmAccountId}
        x = r.get(params=params)
        if r.last_error is None:
        	return x[cls.COLLECTION_NAME]
        else:
        	return r.last_error

class EnvironmentException(BaseException): pass


"""
mixcoatl.admin.billing_code
---------------------------

Implements access to the enStratus Billingcode API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs

import json

class BillingCode(Resource):
    """A billing code is a budget item with optional hard and soft quotas
    against which cloud resources may be provisioned and tracked."""

    PATH = 'admin/BillingCode'
    COLLECTION_NAME = 'billingCodes'
    PRIMARY_KEY = 'billing_code_id'

    def __init__(self, billing_code_id = None, *args, **kwargs):
        Resource.__init__(self)
        self.__billing_code_id = billing_code_id

    @property
    def billing_code_id(self):
        """`int` - The unique id of this billing code"""
        return self.__billing_code_id

    @lazy_property
    def budget_state(self):
        """`str` - The ability of users to provision against this budget"""
        return self.__budget_state

    @lazy_property
    def current_usage(self):
        """`dict` - The month-to-data usage across all clouds for this code"""
        return self.__current_usage

    @lazy_property
    def customer(self):
        """`dict` - The customer to whom this code belongs"""
        return self.__customer

    @lazy_property
    def description(self):
        """`str` - User-friendly description of this code"""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def finance_code(self):
        """`str` - The alphanumeric identifier of this billing code"""
        return self.__finance_code
    
    @finance_code.setter
    def finance_code(self, f):
        self.__finance_code = f

    @lazy_property
    def name(self):
        """`str` - User-friendly name for this billing code"""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def projected_usage(self):
        """`dict` - Estimated end-of-month total to be charged against this budget"""
        return self.__projected_usage

    @lazy_property
    def status(self):
        """`str` - The status of this billing code"""
        return self.__status

    @lazy_property
    def hard_quota(self):
        """`dict` - Cutoff point where no further resources can be billed to this code"""
        return self.__hard_quota

    @hard_quota.setter
    def hard_quota(self, h):
        self.__hard_quota = h

    @lazy_property
    def soft_quota(self):
        """`dict` - Point where budget alerts will be triggered for this billing code"""
        return self.__soft_quota

    @soft_quota.setter
    def soft_quota(self, s):
        self.__soft_quota = s

    @classmethod
    def all(cls, keys_only = False, **kwargs):
        """Get all visible billing codes

        .. note::

            The keys used to make the original request determine result visibility

        :param keys_only: Only return :attr:`billing_code_id` instead of :class:`BillingCode` objects
        :type keys_only: bool.
        :param detail: The level of detail to return - `basic` or `extended`
        :type detail: str.
        :returns: `list` - of :class:`BillingCode` or :attr:`billing_code_id`
        :raises: :class:`BillingCodeException`
        """
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i['billingCodeId'] for i in x[cls.COLLECTION_NAME]]
            else:
                return [cls(i['billingCodeId']) for i in x[cls.COLLECTION_NAME]]
        else:
            raise BillingCodeException(r.last_error)

    @required_attrs(['soft_quota','hard_quota', 'name', 'finance_code', 'description'])
    def add(self):
        """Add a new billing code. """

        payload = { "addBillingCode":[{
                       "softQuota": {"value": self.soft_quota, "currency": "USD"},
                       "hardQuota": {"value": self.hard_quota, "currency": "USD"},
                       "status": "ACTIVE",
                       "name": self.name,
                       "financeCode": self.finance_code,
                       "description": self.description }]}
        response = self.post(data=json.dumps(payload))
        if self.last_error is None:
            return response
        else:
            raise BillingCodeAddException(self.last_error)

    @required_attrs(['billing_code_id'])
    def destroy(self, reason, replacement_code):
        """Destroy billing code with a specified reason :attr:`reason`

        :param reason: The reason of destroying the billing code.
        :type reason: str.
        :param replacement_code: The replacement code.
        :type replacement_code: int.
        :returns: bool -- Result of API call
        """
        p = self.PATH+"/"+str(self.billing_code_id)
        qopts = {'reason':reason, 'replacementCode':replacement_code}
        self.delete(p, params=qopts)
        if self.last_error is None:
            return True
        else:
            raise BillingCodeDestroyException(self.last_error)

class BillingCodeException(BaseException): pass
class BillingCodeAddException(BillingCodeException): pass
class BillingCodeDestroyException(BillingCodeException): pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.utils import camelize, camel_keys
from mixcoatl.admin.job import Job

import json
import time

class Tier(Resource):
    PATH = 'automation/Tier'
    COLLECTION_NAME = 'tiers'
    PRIMARY_KEY = 'tier_id'

    def __init__(self, tier_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__tier_id = tier_id

    @property
    def tier_id(self):
        return self.__tier_id

    @lazy_property
    def budget(self):
        return self.__budget

    @budget.setter
    def budget(self, b):
        # pylint: disable-msg=C0111,W0201
        self.__budget = b

    @lazy_property
    def breach_increment(self):
        return self.__breach_increment

    @breach_increment.setter
    def breach_increment(self, b):
        self.__breach_increment = b

    @lazy_property
    def breach_period_in_minutes(self):
        return self.__breach_period_in_minutes

    @breach_period_in_minutes.setter
    def breach_period_in_minutes(self,b):
        self.__breach_period_in_minutes = b

    @lazy_property
    def cooldown_period_in_minutes(self):
        return self.__cooldown_period_in_minutes

    @cooldown_period_in_minutes.setter
    def cooldown_period_in_minutes_in_minutes(self, m):
        self.__cooldown_period_in_minutes = m

    @lazy_property
    def deployment(self):
        return self.__deployment

    @deployment.setter
    def deployment(self, d):
        # pylint: disable-msg=C0111,W0201
        self.__deployment = d

    @lazy_property
    def label(self):
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def description(self):
        return self.__description

    @description.setter
    def description(self, d):
        # pylint: disable-msg=C0111,W0201
        self.__description = d

    @lazy_property
    def last_breach_change_timestamp(self):
        return self.__last_breach_change_timestamp

    @lazy_property
    def lower_cpu_threshold(self):
        return self.__lower_cpu_threshold

    @lower_cpu_threshold.setter
    def lower_cpu_threshold(self, c):
        self.__lower_cpu_threshold = c

    @lazy_property
    def lower_ram_threshold(self):
        return self.__lower_ram_threshold

    @lower_ram_threshold.setter
    def lower_ram_threshold(self, l):
        self.__lower_ram_threshold = l

    @lazy_property
    def minimum_servers(self):
        return self.__minimum_servers

    @minimum_servers.setter
    def minimum_servers(self, m):
        self.__minimum_servers = m

    @lazy_property
    def maximum_servers(self):
        return self.__maximum_servers

    @maximum_servers.setter
    def maximum_servers(self, m):
        self.__maximum_servers = m

    @lazy_property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        # pylint: disable-msg=C0111,W0201
        self.__name = n

    @lazy_property
    def removable(self):
        return self.__removable

    @lazy_property
    def scaling_rules(self):
        return self.__scaling_rules

    @lazy_property
    def status(self):
        return self.__status

    @lazy_property
    def upper_cpu_threshold(self):
        return self.__upper_cpu_threshold

    @upper_cpu_threshold.setter
    def upper_cpu_threshold(self, u):
        self.__upper_cpu_threshold = u

    @lazy_property
    def upper_ram_threshold(self):
        return self.__upper_ram_threshold

    @upper_ram_threshold.setter
    def upper_ram_threshold(self, u):
        self.__upper_ram_threshold = u

    @classmethod
    def all(cls, **kwargs):
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return r.last_error

    @required_attrs(['name',
										 'description',
										 'budget',
										 'deployment',
										 'minimum_servers',
										 'maximum_servers',
										 'breach_period_in_minutes',
										 'cooldown_period_in_minutes',
										 'lower_cpu_threshold',
										 'upper_cpu_threshold',
										 'lower_ram_threshold',
										 'upper_ram_threshold' ])

    def create(self, callback=None):
        """Creates a new tier

        :param callback: Optional callback to send the resulting :class:`Job`
        :raises: :class:`TierCreationException`
        """

        parms = [{'budget': self.budget,
                    'deployment': {'deploymentId': self.deployment},
                    'description': self.description,
                    'name': self.name,
                    'minimumServers': self.minimum_servers,
                    'maximumServers': self.maximum_servers,
                    'breachIncrement': self.breach_increment,
                    'breachPeriodInMinutes': self.breach_period_in_minutes,
                    'cooldownPeriodInMinutes': self.cooldown_period_in_minutes,
                    'lowerCpuThreshold': self.lower_cpu_threshold,
                    'upperCpuThreshold': self.upper_cpu_threshold,
                    'lowerRamThreshold': self.lower_ram_threshold,
                    'upperRamThreshold': self.upper_ram_threshold}]

        payload = {'addTier':camel_keys(parms)}

        response=self.post(data=json.dumps(payload))
        if self.last_error is None:
            self.load()
            return response
        else:
            raise TierCreationException(self.last_error)

class TierException(BaseException): pass

class TierCreationException(TierException):
    """Tier Creation Exception"""
    pass



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class TierAnalytics(Resource):
    PATH = 'analytics/TierAnalytics'
    COLLECTION_NAME = 'analytics'
    PRIMARY_KEY = 'tier_id'

    def __init__(self, tier_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__tier_id = tier_id

    @property
    def tier_id(self):
        return self.__tier_id

    @lazy_property
    def data_points(self):
        return self.__data_points

    @lazy_property
    def period_start(self):
        return self.__period_start

    @lazy_property
    def period_end(self):
        return self.__period_end

    @lazy_property
    def interval_in_minutes(self):
        return self.__interval_in_minutes

    @classmethod
    def all(cls, tier_id, **kwargs):
        r = Resource(cls.PATH+'/'+str(tier_id))
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return r.last_error



from mixcoatl.resource import Resource
from mixcoatl.admin.job import Job
from mixcoatl.utils import camel_keys
from mixcoatl.decorators.validations import required_attrs
from mixcoatl.decorators.lazy import lazy_property

import json, sys, time

class Server(Resource):
    """A server is a virtual machine running within a data center."""

    PATH = 'infrastructure/Server'
    COLLECTION_NAME = 'servers'
    PRIMARY_KEY = 'server_id'

    def __init__(self, server_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__server_id = server_id

    @property
    def server_id(self):
        """`int` - The enStratus ID of this server"""
        return self.__server_id

    @lazy_property
    def agent_version(self):
        """`int` - The version of the enStratus agent if installed."""
        return self.__agent_version

    @lazy_property
    def cloud(self):
        """`dict` - The cloud provided where the instance is located."""
        return self.__cloud

    @lazy_property
    def label(self):
        """`str` - The label assigned to the server"""
        return self.__label

    @label.setter
    def label(self, l):
        self.__label = l

    @lazy_property
    def customer(self):
        """`dict` - The customer account for the server."""
        return self.__customer

    @lazy_property
    def cm_account(self):
        """`dict` - The configuration management account for the server."""
        return self.__cm_account

    @lazy_property
    def data_center(self):
        """`dict` - The specific datacenter where the instance is located."""
        return self.__data_center

    @data_center.setter
    def data_center(self, d):
        self.__data_center = {u'data_center_id': d}

    @lazy_property
    def cmAccount(self):
        return self.__cmAccount

    @lazy_property
    def environment(self):
        return self.__environment

    @environment.setter
    def environment(self, c):
        self.__environment = {u'sharedEnvironmentCode': c}

    @cmAccount.setter
    def cm_account_id(self, c):
        self.__cmAccount = {u'cmAccountId': c}

    @lazy_property
    def cm_scripts(self):
        return self.__cm_scripts

    @cm_scripts.setter
    def cm_scripts(self, c):
    	s = c.split(",")
    	sc = []
    	for cm in s:
    		sc.append({u'sharedScriptCode': cm})

        self.__cm_scripts = sc

    @lazy_property
    def p_scripts(self):
        return self.__p_scripts

    @p_scripts.setter
    def p_scripts(self, c):
    	s = c.split(",")
    	p = []
    	for cm in s:
    		p.append({'sharedPersonalityCode': cm})

        self.__p_scripts = p

    @lazy_property
    def description(self):
        """The description of the server"""
        return self.__description

    @description.setter
    def description(self, d):
        self.__description = d

    @lazy_property
    def machine_image(self):
        """`dict` - The machine image to use/used to provision the server"""
        return self.__machine_image

    @machine_image.setter
    def machine_image(self, m):
        self.__machine_image = {u'machine_image_id': m}

    @lazy_property
    def vlan(self):
        """`list` - The vlan to assign/assigned to the server"""
        return self.__vlan

    @vlan.setter
    def vlan(self, v):
        self.__vlan = {u'vlan_id': v}

    @lazy_property
    def firewalls(self):
        """`list` - The firewalls to assign/assigned to the server"""
        return self.__firewalls

    @firewalls.setter
    def firewalls(self, f):
        self.__firewalls = f

    @lazy_property
    def name(self):
        """`str` - The name assigned/to assign to the server"""
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n

    @lazy_property
    def owning_groups(self):
        """`list` - The enStratus groups owning the server."""
        return self.__owning_groups

    @lazy_property
    def owning_user(self):
        """`dict` - The enStratus user owning the server."""
        return self.__owning_user

    @lazy_property
    def platform(self):
        """`str` - The platform of the server *(i.e. `UBUNTU`)*."""
        return self.__platform

    @lazy_property
    def personalities(self):
        """`dict` - The personalities associated with the server"""
        return self.__personalities

    @lazy_property
    def private_ip_addresses(self):
        """`list` - The private ip addresses assigned to the server."""

        return self.__private_ip_addresses

    @lazy_property
    def public_ip_address(self):
        """`str` - The public ip address of the server."""
        return self.__public_ip_address

    @lazy_property
    def public_ip_addresses(self):
        """`list` - The list of public ip addresses of the server."""
        return self.__public_ip_addresses

    @lazy_property
    def region(self):
        """`dict` - The region where the server is located"""
        return self.__region

    @lazy_property
    def provider_product_id(self):
        """`str` - The provider's product identifier for the server *(i.e. `m1.large`)*"""
        return self.__provider_product_id

    @provider_product_id.setter
    def provider_product_id(self, id):
        self.__provider_product_id = id

    @lazy_property
    def provider_id(self):
        """`str` - The provider's identifier for the server *(i.e. `i-abcdefg`)*"""
        return self.__provider_id

    @lazy_property
    def product(self):
        """`dict` - The enStratus product record for the server"""
        return self.__product

    @lazy_property
    def start_date(self):
        """`str` - The date the server was started"""
        return self.__start_date

    @lazy_property
    def stop_date(self):
        """`str` - The date the server was stopped"""
        return self.__stop_date

    @lazy_property
    def status(self):
        """`str` - The status of the server *(i.e. `RUNNING` or `PAUSED`)*."""
        return self.__status

    @lazy_property
    def budget(self):
        """`int` - The budget code applied to the server."""
        return self.__budget

    @budget.setter
    def budget(self, b):
        """`int` - The budget code to apply to the server."""
        self.__budget = b

    @lazy_property
    def scripts(self):
        """`list` - The list of configuration management scripts of the server.(Chef?)"""
        return self.__scripts

    @lazy_property
    def run_list(self):
        """`list` - The list of configuration management scripts of the server.(Puppet?)"""
        return self.__run_list

    @lazy_property
    def architecture(self):
        """`str` - The architecture type of the server."""
        return self.__architecture

    @lazy_property
    def terminate_after(self):
        """`str` - The time the server automatically shuts down."""
        return self.__terminate_after

    @terminate_after.setter
    def terminate_after(self, t):
	    self.__terminate_after = t

    @lazy_property
    def pause_after(self):
        """`str` - The time the server automatically pauses."""
        return self.__pause_after



    @property
    def keypair(self):
        """`str` - The keypair to assign

            .. note::
                enStratus does not track keypairs used to launch servers.
                This attribute is used only in the `launch()` call.
        """
        return self.__keypair

    @keypair.setter
    def keypair(self, kp):
        self.__keypair = kp

    def reload(self):
        """Reload resource data from API calls"""
        if self.server_id is not None:
            self.load()
        elif self.current_job is None:
            self.load()
        else:
            if Job.wait_for(self.current_job):
                job = Job(self.current_job)
                self.__server_id = job.message
                self.load()
            else:
                return self.last_error

    @required_attrs(['server_id'])
    def destroy(self, reason='no reason provided'):
        """Terminate server instance with reason :attr:`reason`

        :param reason: The reason for terminating the server
        :type reason: str.
        :returns: bool -- Result of API call
        """
        p = self.PATH+"/"+str(self.server_id)
        qopts = {'reason':reason}
        return self.delete(p, params=qopts)

    @required_attrs(['server_id'])
    def pause(self, reason=None):
        """Pause the server instance with reason :attr:`reason`

        :param reason: The reason for pausing the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'pause':[{}]}

        if reason is not None:
            payload['pause'][0].update({'reason':reason})

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def extend_terminate(self, extend):
        p = '%s/%s' % (self.PATH, str(self.server_id))
        qopts = {'terminateAfter':extend}
        return self.delete(p, params=qopts)

    @required_attrs(['server_id'])
    def start(self, reason=None):
        """Start the paused server instance with reason :attr:`reason`

        :param reason: The reason for starting the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'start':[{}]}

        if reason is not None:
            payload['start'][0].update({'reason':reason})

        return self.put(p, data=json.dumps(payload))

    @required_attrs(['server_id'])
    def stop(self, reason=None):
        """Stop the server instance with reason :attr:`reason`

        :param reason: The reason for stopping the server
        :type reason: str.
        :returns: Job -- Result of API call
        """
        p = '%s/%s' % (self.PATH, str(self.server_id))
        payload = {'stop':[{}]}

        if reason is not None:
            payload['stop'][0].update({'reason':reason})

        return self.put(p, data=json.dumps(payload))

    # TODO: Refactor this a bit. We should be raising exceptions instead of
    # this madness of returning the last error. Makes no sense. I should have
    # never done it.
    @required_attrs(['provider_product_id', 'machine_image', 'description',
                    'name', 'data_center', 'budget'])
    def launch(self, callback=None):
        """Launches a server with the configured parameters

        >>> def cb(j): print(j)
        >>> s = Server()
        >>> s.provider_product_id = 'm1.large'
        >>> s.machine_image = 12345
        >>> s.description = 'my first launch'
        >>> s.name = 'server-1-test'
        >>> s.data_center = 54321
        >>> s.keypair = 'my-aws-keypair'
        >>> s.launch(callback=cb)

        :param callback: Optional callback to send the results of the API call
        :type callback: func.
        :returns: int -- The job id of the launch request
        :raises: :class:`ServerLaunchException`, :class:`mixcoatl.decorators.validations.ValidationException`
        """
        optional_attrs = ['vlan', 'firewalls', 'keypair', 'label', 'cmAccount', 'environment', 'cm_scripts', 'p_scripts', 'volumeConfiguration']
        if self.server_id is not None:
            raise ServerLaunchException('Cannot launch an already running server: %s' % self.server_id)

        payload = {'launch':
                    [{
                        'productId': self.provider_product_id,
                        'budget': self.budget,
                        'machineImage': camel_keys(self.machine_image),
                        'description': self.description,
                        'name': self.name,
                        'dataCenter': camel_keys(self.data_center)
                    }]}

        for oa in optional_attrs:
            try:
                if getattr(self, oa) is not None:
                    if oa == 'cm_scripts':
                        payload['launch'][0].update({'scripts':getattr(self, oa)})
                    elif oa == 'p_scripts':
                        payload['launch'][0].update({'personalities':getattr(self, oa)})
                    elif oa == 'volumeConfiguration':
                        payload['launch'][0].update({'volumeConfiguration':{u'raidlevel':'RAID0', u'volumeCount':1, u'volumeSize':2, u'fileSystem':'ext3', u'mountPoint':'/mnt/data'}})
                    elif oa == 'vlan':
                        payload['launch'][0].update({'vlan':camel_keys(getattr(self, oa))})
                    else:
                        payload['launch'][0].update({oa:getattr(self, oa)})
            except AttributeError:
                pass

        self.post(data=json.dumps(payload))
        if self.last_error is None:
            if callback is not None:
                callback(self.current_job)
            else:
                return self.current_job
        else:
            raise ServerLaunchException(self.last_error)

    def duplicate(self, server):
        pass

    def wait_for(self, status='RUNNING', callback = None):
        """Blocks execution until the current server has status of :attr:`status`

        :param status: The status to expect before continuing *(i.e. `RUNNING` or `PAUSED`)*
        :type status: str.
        :param callback: Optional callback to be called with the final :class:`Server`` when ``status`` is reached
        :type callback: func.
        :raises: `ServerException`
        """
        if self.server_id is None:
            raise ServerException('Must be called with an existing server.')
        initial_status = self.status
        if self.last_error is None:
            if initial_status == status:
                return self
            while self.status != status:
                time.sleep(5)
                self.load()
                if self.last_error is not None:
                    raise ServerException(self.last_error)
                else:
                    continue
            if callback is not None:
                callback(self)
            else:
                return self

    @classmethod
    def all(cls, **kwargs):
        """Get a list of all known servers

        >>> Server.all()
        [{'server_id':1,...},{'server_id':2,...}]

        :returns: list -- a list of :class:`Server`
        :raises: ServerException
        """
        r = Resource(cls.PATH)
        r.request_details = 'basic'

        if 'params' in kwargs:
          params = kwargs['params']
        else:
          params = []
        s = r.get(params=params)
        if r.last_error is None:
            servers = [cls(server['serverId']) for server in s[cls.COLLECTION_NAME]]
            return servers
        else:
            raise ServerException(r.last_error)

class ServerException(BaseException): pass
class ServerLaunchException(ServerException): pass



"""
mixcoatl.admin.job
------------------

Implements access to the enStratus Job API
"""
from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

import time

class Job(Resource):
    PATH = 'admin/Job'
    COLLECTION_NAME = 'jobs'
    PRIMARY_KEY = 'job_id'

    def __init__(self, job_id = None, **kwargs):
        Resource.__init__(self)
        self.__job_id = job_id

    @property
    def job_id(self):
        """`int` The unique enStratus id for this job"""
        return self.__job_id

    @lazy_property
    def status(self):
        """`str` The current status of this job"""
        return self.__status

    @lazy_property
    def description(self):
        """`str` The description of the running job."""
        return self.__description

    @lazy_property
    def message(self):
        """`str` A message describing the current disposition of the operation"""
        return self.__message

    @lazy_property
    def start_date(self):
        """`str` The date and time when the job was started"""
        return self.__start_date

    @lazy_property
    def end_date(self):
        """`str` The data and time when the job was completed
        `None` if :attr:`status is `RUNNING`
        """
        return self.__end_date

    @classmethod
    def all(cls, keys_only=False):
        """Get all jobs


        :param keys_only: Only return :attr:`job_id` instead of :class:`Job`
        :type keys_only: bool.
        :returns: `list` of :class:`Job` or :attr:`job_id`
        :raises: :class:`JobException`
        """
        r = Resource(cls.PATH)
        x = r.get()
        if r.last_error is None:
            if keys_only is True:
                return [i[camelize(cls.PRIMARY_KEY)] for i in x[cls.COLLECTION_NAME]]
            else:
                return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            raise JobException(r.last_error)

    @classmethod
    def wait_for(cls, job_id, status='COMPLETE', callback = None):
        """Blocks execution until :attr:`job_id` returns :attr:`status`

        :param job_id: The ID of the job to wait on
        :type job_id: int.
        :param status: The status to expect before continuing
        :type status: str.
        :param callback: Optional callback to be called with the final :class:`Job`
            when :attr:`status` is reached
        :type callback: func.
        :returns: `bool` - Result of job exectution
        :raises: :class:`JobException`
        """
        j = Job(job_id)
        j.load()
        if j.last_error is not None:
            raise JobException(j.last_error)
        else:
            while j.status != status:
                time.sleep(5)
                j.load()
                if j.status == 'ERROR':
                    break
                else:
                    continue
        if callback is not None:
            callback(j)
        else:
            return True

class JobException(BaseException): pass



from mixcoatl.admin.billing_code import BillingCode
from mixcoatl.geography.region import Region
from mixcoatl.admin.group import Group
from mixcoatl.admin.user import User

def get_servers(servers, **kwargs):
    """ Returns a list of servers

    Arguments:
    :param servers: a list of servers that needs to be filtered.

    Keyword arguments:
    :param account_user_id: owning user's account user ID.
    :param vm_login_id: owning user's VM login ID.
    :param email: owning user's email address.
    :param group_id: owning group's group ID.
    :param budget_id: budget ID.
    :returns: a list of filtered servers.
    :rtype: list
    """
    filtered_servers = servers

    if kwargs.has_key('account_user_id') and kwargs['account_user_id'] is not None:
        filtered_servers = [server for server in servers if hasattr(server, 'owning_user') and
                            server.owning_user.has_key('account_user_id') and
                            server.owning_user['account_user_id'] == kwargs['account_user_id']]
    if kwargs.has_key('vm_login_id') and kwargs['vm_login_id'] is not None:
        if filtered_servers is not None: servers = filtered_servers
        filtered_servers = [server for server in servers if hasattr(server, 'owning_user') and
                            server.owning_user.has_key('vm_login_id') and
                            server.owning_user['vm_login_id'] == kwargs['vm_login_id']]
    if kwargs.has_key('email') and kwargs['email'] is not None:
        if filtered_servers is not None: servers = filtered_servers
        filtered_servers = [server for server in servers if hasattr(server, 'owning_user') and
                            server.owning_user.has_key('email') and
                            server.owning_user['email'] == kwargs['email']]
    if kwargs.has_key('group_id') and kwargs['group_id'] is not None:
        if filtered_servers is not None: servers = filtered_servers
        filtered_servers = [server for server in servers if hasattr(server, 'owning_groups')
                            for group in server.owning_groups if group['group_id'] == int(kwargs['group_id'])]
    if kwargs.has_key('budget_id') and kwargs['budget_id'] is not None:
        if filtered_servers is not None: servers = filtered_servers
        filtered_servers = [server for server in servers if hasattr(server, 'budget') and
                            server.budget == int(kwargs['budget_id'])]

    return filtered_servers

def get_snapshots(snapshots, **kwargs):
    """ Returns a list of snapshots

    Arguments:
    :param snapshots: a list of snapshots that needs to be filtered.

    Keyword arguments:
    :param group_id: owning group's group ID.
    :param budget_id: budget ID.
    :returns: a list of filtered snapshots.
    :rtype: list
    """
    filtered_snapshots = snapshots

    if kwargs.has_key('group_id') and kwargs['group_id'] is not None:
        filtered_snapshots = [snapshot for snapshot in snapshots if hasattr(snapshot, 'owning_groups')
                              for g in snapshot.owning_groups if g['group_id'] == int(kwargs['group_id'])]
    if kwargs.has_key('budget_id') and kwargs['budget_id'] is not None:
        if filtered_snapshots is not None: snapshots = filtered_snapshots
        filtered_snapshots = [snapshot for snapshot in snapshots if hasattr(snapshot, 'budget') and
                              snapshot.budget == int(kwargs['budget_id'])]

    return filtered_snapshots

def get_volumes(volumes, **kwargs):
    """ Returns a list of volumes

    Arguments:
    :param volumes: a list of volumes that needs to be filtered.

    Keyword arguments:
    :param vm_login_id: owning user's VM login ID.
    :param email: owning user's email address.
    :param group_id: owning group's group ID.
    :param budget_id: budget ID.
    :param size: minimum size of the volume.
    :returns: a list of filtered volumes.
    :rtype: list
    """
    filtered_volumes = volumes

    if kwargs.has_key('vm_login_id') and kwargs['vm_login_id'] is not None:
        filtered_volumes = [volume for volume in volumes if hasattr(volume, 'owning_user') and
                            volume.owning_user.has_key('vm_login_id') and
                            volume.owning_user['vm_login_id'] == kwargs['vm_login_id']]
    if kwargs.has_key('email') and kwargs['email'] is not None:
        if filtered_volumes is not None: volumes = filtered_volumes
        filtered_volumes = [volume for volume in volumes if hasattr(volume, 'owning_user') and
                            volume.owning_user.has_key('email') and
                            volume.owning_user['email'] == kwargs['email']]
    if kwargs.has_key('group_id') and kwargs['group_id'] is not None:
        if filtered_volumes is not None: volumes = filtered_volumes
        filtered_volumes = [volume for volume in volumes if hasattr(volume, 'owning_groups')
                            for group in volume.owning_groups if group['group_id'] == int(kwargs['group_id'])]
    if kwargs.has_key('budget_id') and kwargs['budget_id'] is not None:
        if filtered_volumes is not None: volumes = filtered_volumes
        filtered_volumes = [volume for volume in volumes if hasattr(volume, 'budget') and
                            volume.budget == int(kwargs['budget_id'])]
    if kwargs.has_key('size') and kwargs['size'] is not None:
        if filtered_volumes is not None: volumes = filtered_volumes
        filtered_volumes = [volume for volume in volumes if volume.size_in_gb >= int(kwargs['size'])]

    return filtered_volumes

def get_user(users, **kwargs):
    """ Returns a user that matches with arguments.

    Arguments:
    :param users: a list of users that needs to be filtered.

    Keyword arguments:
    :param vm_login_id: owning user's VM login ID.
    :param email: owning user's email address.
    :returns: a list of filtered users.
    :rtype: list
    """
    selected_user = users

    if kwargs.has_key('vm_login_id') and kwargs['vm_login_id'] is not None:
        for user in users:
            if hasattr(user, 'vm_login_id') and user.vm_login_id == kwargs['vm_login_id']:
                selected_user = user
    elif kwargs.has_key('email') and kwargs['email'] is not None:
        for user in users:
            if hasattr(user, 'email') and user.email == kwargs['email']:
                selected_user = user

    return selected_user

def get_account_user_id(**kwargs):
    """ Returns account_user_id from arguments

    Keyword arguments:
    :param vm_login_id: user's VM login ID like p100
    :param email: user's E-Mail address
    :returns: account_user_id
    :rtype: int
    """

    if 'vm_login_id' in kwargs:
        users = User.all()
        selected_user = get_user(users, vm_login_id=kwargs['vm_login_id'])
    elif 'email' in kwargs:
        users = User.all()
        selected_user = get_user(users, email=kwargs['email'])

    return selected_user.account_user_id

def get_vm_login_id(**kwargs):
    """ Returns vm_login_id from arguments

    Keyword arguments:
    :param email: user's E-Mail address
    :returns: vm_login_id
    :rtype: str
    """

    if 'email' in kwargs:
        users = User.all()
        selected_user = get_user(users, email=kwargs['email'])

    return selected_user.vm_login_id

def get_budget_id(budget_name):
    """ Returns budget_id from arguments

    Arguments:
    :param budget_name: budget name
    :returns: budget_id
    :rtype: int
    """
    budgets = BillingCode.all(detail='basic')

    for budget in budgets:
        if hasattr(budget, 'name') and budget.name == budget_name:
            selected_budget = budget

    return selected_budget.billing_code_id

def get_group_id(group_name):
    """ Returns a group ID from group name

    Arguments:
    :param group_name: name of the group
    :returns: group_id
    :rtype: int
    """
    groups = Group.all(detail='basic')

    for group in groups:
        if hasattr(group, 'name') and group.name == group_name:
            selected_group = group

    return selected_group.group_id

def get_region_id(region_pid):
    """ Returns a region ID from provider_id such as us-east-1.

    Arguments:
    :param region_pid: provider ID of the region such as us-east-1
    :returns: region_id such as 19343
    :rtype: int
    """
    regions = Region.all(detail='basic')

    for region in regions:
        if hasattr(region, 'provider_id') and region.provider_id == region_pid:
            selected_region = region

    return selected_region.region_id



from mixcoatl.resource import Resource
from mixcoatl.decorators.lazy import lazy_property
from mixcoatl.utils import camelize

class Service(Resource):
    PATH = 'automation/Service'
    COLLECTION_NAME = 'services'
    PRIMARY_KEY = 'service_id'

    def __init__(self, service_id=None, *args, **kwargs):
        Resource.__init__(self)
        self.__service_id = service_id

    @property
    def service_id(self):
        return self.__service_id

    @lazy_property
    def backup_interval_in_minutes(self):
        return self.__backup_interval_in_minutes

    @lazy_property
    def budget(self):
        return self.__budget

    @lazy_property
    def description(self):
        return self.__description

    @lazy_property
    def name(self):
        return self.__name

    @lazy_property
    def owning_group(self):
        return self.__owning_group

    @lazy_property
    def owning_user(self):
        return self.__owning_user

    @lazy_property
    def run_as_user(self):
        return self.__run_as_user

    @lazy_property
    def scaling_model(self):
        return self.__scaling_model

    @lazy_property
    def status(self):
        return self.__status

    @classmethod
    def all(cls, **kwargs):
        r = Resource(cls.PATH)
        if 'details' in kwargs:
            r.request_details = kwargs['details']
        else:
            r.request_details = 'basic'

        x = r.get()
        if r.last_error is None:
            return [cls(i[camelize(cls.PRIMARY_KEY)]) for i in x[cls.COLLECTION_NAME]]
        else:
            return r.last_error
